-- ---------------------------------------------------------------------
    -- ddl_util
    -- Copyright 2018: Ralf Beckmann
    -- V. 1.0.0
    --
    -- https://informatik-transparent.de
    -- ---------------------------------------------------------------------
    -- Der Autor der Beitrge, Programme , Downloads, Quellcodes und Plug/Ins 
    -- haftet nicht fr Schden an Soft- und Hardware oder Vermgensschden, 
    -- die durch das Benutzen entstehen, es sei denn diese beruhen auf grob 
    -- fahrlssigen oder vorstzlichen Handeln des Autors, seiner Erfllungsgehilfen 
    -- oder seiner gesetzlichen Vertreter.
    --
    -- Bei den Beitrgen handelt es sich um eine Beispiele, die nur zu 
    -- Anschauungszwecken gedacht sind. Bitte Verwenden Sie die Inhalte 
    -- nur in einem Testsystem. Sie sind selber fr die Qualittssicherung 
    -- und die Nutzung verantwortlich. 
    -- Es drfen keine Schutzvermerke entfernt werden.
    -- ---------------------------------------------------------------------
/*
 DROP TRIGGER trg_ddl_util_before;
DROP TRIGGER trg_ddl_util_after;
DROP PACKAGE ddl_util;
DROP TABLE ddl_object_state;
DROP TABLE ddl_object_state_hist; */
CREATE TABLE ddl_object_state (
    object_type    VARCHAR2(50 BYTE),
    object_owner   VARCHAR2(32 BYTE),
    object_name    VARCHAR2(32 BYTE), 
    changed_on     DATE,
    changed_by     VARCHAR2(32 BYTE)
)
/

CREATE TABLE ddl_object_state_hist (
    object_type    VARCHAR2(50 BYTE),
    object_owner   VARCHAR2(32 BYTE),
    object_name    VARCHAR2(32 BYTE),
    status_code    VARCHAR2(20 BYTE),
    changed_on     DATE,
    changed_by     VARCHAR2(32 BYTE),
    ddl_sql        CLOB
)
/

--------------------------------------------------------
--  DDL for Package DDL_UTIL
--------------------------------------------------------

CREATE OR REPLACE PACKAGE ddl_util IS
    -- ---------------------------------------------------------------------
    -- ddl_util
    -- Copyright 2018: Ralf Beckmann
    --
    -- https://informatik-transparent.de
    -- ---------------------------------------------------------------------
PROCEDURE check_object_state (
        p_object_type   VARCHAR2,
        p_owner         VARCHAR2,
        p_name          VARCHAR2
    );

    PROCEDURE lock_object (
        p_object_type   VARCHAR2,
        p_owner         VARCHAR2,
        p_name          VARCHAR2,
        p_trace_hist    BOOLEAN DEFAULT true
    );

    PROCEDURE unlock_object (
        p_object_type   VARCHAR2,
        p_owner         VARCHAR2,
        p_name          VARCHAR2,
        p_trace_hist    BOOLEAN DEFAULT true,
        p_secure_mode   BOOLEAN DEFAULT true
    );

END;                                                           -- Package spec

/
CREATE OR REPLACE PACKAGE BODY ddl_util IS
    -- ---------------------------------------------------------------------
    -- ddl_util
    -- Copyright 2018: Ralf Beckmann
    --
    -- https://informatik-transparent.de
    -- ---------------------------------------------------------------------

    PROCEDURE check_object_state (
        p_object_type   VARCHAR2,
        p_owner         VARCHAR2,
        p_name          VARCHAR2
    ) IS
        l_c            NUMBER;
        l_locked_by    VARCHAR2(32);
        l_changed_on   DATE;
    BEGIN
        SELECT
            COUNT(*)
        INTO l_c
        FROM
            ddl_object_state
        WHERE
            object_owner = p_owner
            AND object_name = p_name;

        IF
            l_c = 0
        THEN
            raise_application_error(-20003,'Please lock the object before alter');
        END IF;
        SELECT
            changed_by,
            changed_on
        INTO
            l_locked_by,
            l_changed_on
        FROM
            ddl_object_state
        WHERE
            object_owner = p_owner
            AND object_name = p_name;

        IF
            l_locked_by <> ora_login_user
        THEN
            raise_application_error(-20001,'Object was already locked by '
                                            || l_locked_by
                                            || ' at '
                                            || TO_CHAR(l_changed_on,'dd.mm.yyyy HH24:Mi:ss') );
        END IF;

    END;

    PROCEDURE lock_object (
        p_object_type   VARCHAR2,
        p_owner         VARCHAR2,
        p_name          VARCHAR2,
        p_trace_hist    BOOLEAN DEFAULT true
    ) IS
        l_c             NUMBER;
        l_locked_by     VARCHAR2(32);
        l_changed_on    DATE;
        l_ddl_extract   CLOB;
    BEGIN
        SELECT
            COUNT(*)
        INTO l_c
        FROM
            ddl_object_state
        WHERE
            object_owner = p_owner
            AND object_name = p_name;

        IF
            l_c = 0
        THEN
            INSERT INTO ddl_object_state (
                object_type,
                object_owner,
                object_name,
                changed_on,
                changed_by
            )
                SELECT
                    p_object_type,
                    p_owner object_owner,
                    p_name object_name,
                    SYSDATE changed_on,
                    ora_login_user changed_by
                FROM
                    dual;

            IF
                p_trace_hist
            THEN
                l_ddl_extract := dbms_metadata.get_ddl(p_object_type,p_name);
                INSERT INTO ddl_object_state_hist (
                    object_type,
                    object_owner,
                    object_name,
                    ddl_sql,
                    status_code,
                    changed_on,
                    changed_by
                )
                    SELECT
                        p_object_type,
                        p_owner object_owner,
                        p_name object_name,
                        l_ddl_extract,
                        'LOCK',
                        SYSDATE changed_on,
                        ora_login_user changed_by
                    FROM
                        dual;

            END IF;

        ELSE
            SELECT
                changed_by,
                changed_on
            INTO
                l_locked_by,
                l_changed_on
            FROM
                ddl_object_state
            WHERE
                object_owner = p_owner
                AND object_name = p_name;

            IF
                l_locked_by <> ora_login_user
            THEN
                raise_application_error(-20001,'Object was already locked by '
                                                || l_locked_by
                                                || ' at '
                                                || TO_CHAR(l_changed_on,'dd.mm.yyyy HH24:Mi:ss') );

            END IF;

        END IF;

    END;

    PROCEDURE unlock_object (
        p_object_type   VARCHAR2,
        p_owner         VARCHAR2,
        p_name          VARCHAR2,
        p_trace_hist    BOOLEAN DEFAULT true,
        p_secure_mode   BOOLEAN DEFAULT true
    ) IS
        l_c             NUMBER;
        l_locked_by     VARCHAR2(32);
        l_changed_on    DATE;
        l_ddl_extract   CLOB;
    BEGIN
        SELECT
            COUNT(*)
        INTO l_c
        FROM
            ddl_object_state
        WHERE
            object_owner = p_owner
            AND object_name = p_name;

        IF
            l_c = 1
        THEN
            SELECT
                changed_by,
                changed_on
            INTO
                l_locked_by,
                l_changed_on
            FROM
                ddl_object_state
            WHERE
                object_owner = p_owner
                AND object_name = p_name;

            dbms_output.put_line(substr('Value of l_locked_by=' || TO_CHAR(l_locked_by),1,255) );

            IF
                p_secure_mode AND l_locked_by <> ora_login_user
            THEN
                raise_application_error(-20002,'You have no permissions in secure mode to unlock the object. The Object was locked by '
                                                || l_locked_by
                                                || ' at '
                                                || TO_CHAR(l_changed_on,'dd.mm.yyyy HH24:Mi:ss') );
            END IF;

            DELETE ddl_object_state
            WHERE
                object_owner = p_owner
                AND object_name = p_name;

            IF
                p_trace_hist
            THEN
                l_ddl_extract := dbms_metadata.get_ddl(p_object_type,p_name);
                INSERT INTO ddl_object_state_hist (
                    object_type,
                    object_owner,
                    object_name,
                    ddl_sql,
                    status_code,
                    changed_on,
                    changed_by
                )
                    SELECT
                        p_object_type,
                        p_owner object_owner,
                        p_name object_name,
                        l_ddl_extract,
                        'UNLOCK',
                        SYSDATE changed_on,
                        ora_login_user changed_by
                    FROM
                        dual;

            END IF;

        END IF;

    END;

END;
/ 

CREATE OR REPLACE TRIGGER trg_ddl_util_before BEFORE ALTER OR DROP ON DATABASE DECLARE BEGIN
    IF
        ora_dict_obj_name NOT IN (
            'TRG_DDL_UTIL_BEFORE',
            'TRG_DDL_UTIL_AFTER',
            'DDL_UTIL'
        ) AND ora_login_user NOT IN (
            'SYS',
            'SYSTEM'
        )
    THEN
        ddl_util.lock_object(ora_dict_obj_type,ora_dict_obj_owner,ora_dict_obj_name);
    END IF;
END;
/

CREATE OR REPLACE TRIGGER trg_ddl_util_after AFTER CREATE OR DROP ON DATABASE DECLARE BEGIN
    IF
        ora_dict_obj_name NOT IN (
            'TRG_DDL_UTIL_BEFORE',
            'TRG_DDL_UTIL_AFTER',
            'DDL_UTIL'
        ) AND ora_login_user NOT IN (
            'SYS',
            'SYSTEM'
        )
    THEN
        IF
            ora_sysevent = 'CREATE'
        THEN
            ddl_util.lock_object(ora_dict_obj_type,ora_dict_obj_owner,ora_dict_obj_name);
        END IF;

        IF
            ora_sysevent = 'DROP'
        THEN
            ddl_util.unlock_object(ora_dict_obj_type,ora_dict_obj_owner,ora_dict_obj_name,false,true);
        END IF;

    END IF;
END;
/